package me.axieum.mcmod.authme.api.gui.widget;

import org.jetbrains.annotations.Nullable;

import com.mojang.blaze3d.systems.RenderSystem;
import me.axieum.mcmod.authme.api.util.SessionUtils;
import me.axieum.mcmod.authme.api.util.SessionUtils.SessionStatus;
import me.axieum.mcmod.authme.impl.AuthMe;
import net.minecraft.class_2561;
import net.minecraft.class_2588;
import net.minecraft.class_310;
import net.minecraft.class_344;
import net.minecraft.class_4185;
import net.minecraft.class_437;
import net.minecraft.class_4587;

/**
 * The textured button widget for opening the authentication screens.
 */
public class AuthButtonWidget extends class_344
{
    // The screen used to constrain the button movements to
    private final @Nullable class_437 screen;
    // Callback for after the button has been dragged
    private final @Nullable MoveAction moveAction;
    // True if the button was just dragged
    private boolean didDrag = false;
    // The last known status of the Minecraft session
    private SessionStatus sessionStatus = SessionStatus.UNKNOWN;

    /**
     * Constructs a fixed (no drag) authentication button.
     *
     * @param x           x coordinate
     * @param y           y coordinate
     * @param pressAction on click action
     * @see #AuthButtonWidget(class_437, int, int, class_4241, MoveAction)
     */
    public AuthButtonWidget(int x, int y, class_4241 pressAction)
    {
        this(null, x, y, pressAction, null);
    }

    /**
     * Constructs a fixed (no drag) authentication button with non-visible text.
     *
     * @param x           x coordinate
     * @param y           y coordinate
     * @param pressAction on click action
     * @param message     non-visible button text
     * @see #AuthButtonWidget(class_437, int, int, class_4241, MoveAction, class_5316, class_2561)
     */
    public AuthButtonWidget(int x, int y, class_4241 pressAction, @Nullable class_2561 message)
    {
        this(null, x, y, pressAction, null, null, message);
    }

    /**
     * Constructs a fixed (no drag) authentication button with non-visible text and a tooltip.
     *
     * @param x               x coordinate
     * @param y               y coordinate
     * @param pressAction     on click action
     * @param tooltipSupplier tooltip supplier
     * @param message         non-visible button text
     * @see #AuthButtonWidget(class_437, int, int, class_4241, MoveAction, class_5316, class_2561)
     */
    public AuthButtonWidget(
        int x, int y, class_4241 pressAction, @Nullable class_5316 tooltipSupplier, @Nullable class_2561 message
    )
    {
        this(null, x, y, pressAction, null, tooltipSupplier, message);
    }

    /**
     * Constructs a movable (with drag) authentication button.
     *
     * @param screen      screen to constrain movement to
     * @param x           initial x coordinate
     * @param y           initial y coordinate
     * @param pressAction on click action
     * @param moveAction  on move action
     * @see #AuthButtonWidget(class_437, int, int, class_4241, MoveAction, class_2561)
     */
    public AuthButtonWidget(
        @Nullable class_437 screen, int x, int y, class_4241 pressAction, @Nullable MoveAction moveAction
    )
    {
        this(screen, x, y, pressAction, moveAction, new class_2588("gui.authme.button.auth"));
    }

    /**
     * Constructs a movable (with drag) authentication button with non-visible text.
     *
     * @param screen      screen to constrain movement to
     * @param x           initial x coordinate
     * @param y           initial y coordinate
     * @param pressAction on click action
     * @param moveAction  on move action
     * @param message     non-visible button text
     * @see #AuthButtonWidget(class_437, int, int, class_4241, MoveAction, class_5316, class_2561)
     */
    public AuthButtonWidget(
        @Nullable class_437 screen,
        int x,
        int y,
        class_4241 pressAction,
        @Nullable MoveAction moveAction,
        @Nullable class_2561 message
    )
    {
        this(screen, x, y, pressAction, moveAction, null, message);
    }

    /**
     * Constructs a movable (with drag) authentication button with non-visible text and a tooltip.
     *
     * @param screen          screen to constrain movement to
     * @param x               initial x coordinate
     * @param y               initial y coordinate
     * @param pressAction     on click action
     * @param moveAction      on move action
     * @param tooltipSupplier tooltip supplier
     * @param message         non-visible button text
     */
    public AuthButtonWidget(
        @Nullable class_437 screen,
        int x,
        int y,
        class_4241 pressAction,
        @Nullable MoveAction moveAction,
        @Nullable class_5316 tooltipSupplier,
        class_2561 message
    )
    {
        super(x, y, 20, 20, 0, 146, 20, class_4185.field_22757, 256, 256, pressAction, tooltipSupplier, message);
        this.screen = screen;
        this.moveAction = moveAction;
        refreshSessionStatus();
    }

    /**
     * Returns the last known Minecraft session status.
     *
     * @return cached Minecraft session status
     */
    public SessionStatus getSessionStatus()
    {
        return sessionStatus;
    }

    /**
     * Refreshes the Minecraft session status.
     */
    public void refreshSessionStatus()
    {
        sessionStatus = SessionStatus.UNKNOWN;
        SessionUtils.getStatus().thenAccept(status -> sessionStatus = status);
    }

    /**
     * Sets the Minecraft session status.
     *
     * @param sessionStatus Minecraft session status
     */
    public void setSessionStatus(SessionStatus sessionStatus)
    {
        this.sessionStatus = sessionStatus;
    }

    /**
     * Adjusts the default mouse down behaviour to not trigger a click - rather
     * handle clicks on mouse up/release. This allows the user to initially
     * click to begin dragging the button to a new position.
     *
     * @see #method_25406 for super invocation
     */
    @Override
    public boolean method_25402(double mouseX, double mouseY, int button)
    {
        if (this.moveAction != null) {
            return this.method_25351(button) && this.method_25361(mouseX, mouseY);
        }
        return super.method_25402(mouseX, mouseY, button);
    }

    /**
     * Adjusts the default mouse up behaviour to trigger a click.
     *
     * @see super#mouseClicked for mouse click implementation
     */
    @Override
    public boolean method_25406(double mouseX, double mouseY, int button)
    {
        if (this.moveAction == null) return super.method_25406(mouseX, mouseY, button);
        if (!this.method_25351(button)) return false;
        // Check if the user just finished dragging
        if (this.didDrag) {
            // Invoke any callbacks and reset the dragging status
            this.moveAction.onMove(this);
            return this.didDrag = false;
        }
        // Else, continue as a button press
        return super.method_25402(mouseX, mouseY, button);
    }

    /**
     * Allows the button to be moved while dragging.
     *
     * @see #method_25406 for where dragging is finalised
     */
    @Override
    protected void method_25349(double mouseX, double mouseY, double deltaX, double deltaY)
    {
        if (this.moveAction != null) {
            // Flag the widget as in a dragging state
            this.didDrag = true;

            // Move the button with the drag, constraining within the screen's bounds
            if (this.screen != null) {
                this.method_1893(
                    Math.min(Math.max(0, (int) mouseX - this.field_22758 / 2), this.screen.field_22789 - this.field_22758),
                    Math.min(Math.max(0, (int) mouseY - this.field_22759 / 2), this.screen.field_22790 - this.field_22759)
                );
            } else {
                this.method_1893((int) mouseX - this.field_22758 / 2, (int) mouseY - this.field_22759 / 2);
            }
        }

        super.method_25349(mouseX, mouseY, deltaX, deltaY);
    }

    @Override
    public void method_25359(class_4587 matrices, int mouseX, int mouseY, float delta)
    {
        // Cascade the rendering
        super.method_25359(matrices, mouseX, mouseY, delta);

        // Render the current session status
        class_310.method_1551().method_1531().method_22813(AuthMe.WIDGETS_TEXTURE);
        RenderSystem.enableDepthTest();
        final int u;
        switch (sessionStatus) {
            case VALID:
                u = 0;
                break;
            case OFFLINE:
                u = 8;
                break;
            default:
                u = 16;
        }
        method_25290(matrices, field_22760 + field_22758 - 6, field_22761 - 1, u, 60, 8, 8, 128, 128);

        // Render the tooltip on top of the session status
        if (method_25367()) method_25352(matrices, mouseX, mouseY);
    }

    /**
     * A functional interface for widget relocation callback.
     */
    @FunctionalInterface
    public interface MoveAction
    {
        /**
         * Callback for when the widget has been dragged elsewhere.
         *
         * @param button button that was moved
         */
        void onMove(class_4185 button);
    }
}
